from abc import ABC, abstractmethod
from pathlib import Path


class BlobBackend(ABC):
    """A blob backend downloads large remote files."""

    @abstractmethod
    def download(self, url: str, download_to: Path):
        raise NotImplementedError()



import inspect
from configparser import DEFAULTSECT, SafeConfigParser
from functools import partial
from operator import is_not
from pathlib import Path
from typing import Dict, List, Type

from truss.remote.baseten import BasetenRemote
from truss.remote.truss_remote import RemoteConfig, TrussRemote

USER_TRUSSRC_PATH = Path("~/.trussrc").expanduser()


def load_config() -> SafeConfigParser:
    config = SafeConfigParser()
    config.read(USER_TRUSSRC_PATH)
    return config


def update_config(config: SafeConfigParser):
    with open(USER_TRUSSRC_PATH, "w") as configfile:
        config.write(configfile)


class RemoteFactory:
    """
    A factory for instantiating a TrussRemote from a .trussrc file and a user-specified remote config name
    """

    REGISTRY: Dict[str, Type[TrussRemote]] = {"baseten": BasetenRemote}

    @staticmethod
    def get_available_config_names() -> List[str]:
        if not USER_TRUSSRC_PATH.exists():
            return []

        config = load_config()
        return list(filter(partial(is_not, DEFAULTSECT), config.keys()))

    @staticmethod
    def load_remote_config(remote_name: str) -> RemoteConfig:
        """
        Load and validate a remote config from the .trussrc file
        """
        if not USER_TRUSSRC_PATH.exists():
            raise FileNotFoundError("No ~/.trussrc file found.")

        config = load_config()

        if remote_name not in config:
            raise ValueError(f"Service provider {remote_name} not found in ~/.trussrc")

        return RemoteConfig(name=remote_name, configs=dict(config[remote_name]))

    @staticmethod
    def update_remote_config(remote_config: RemoteConfig):
        """
        Load and validate a remote config from the .trussrc file
        """
        config = load_config()
        config[remote_config.name] = remote_config.configs
        update_config(config)

    @staticmethod
    def validate_remote_config(remote_config: Dict, remote_name: str):
        """
        Validates remote config by checking
            1. the 'remote' field exists
            2. all required parameters for the 'remote' class are provided
        """
        if "remote_provider" not in remote_config:
            raise ValueError(
                f"Missing 'remote_provider' field for remote {remote_name} in .trussrc"
            )

        if remote_config["remote_provider"] not in RemoteFactory.REGISTRY:
            raise ValueError(
                f"Remote provider {remote_config['remote_provider']} not found in registry"
            )

        remote = RemoteFactory.REGISTRY.get(remote_config["remote_provider"])
        if remote:
            required_params = RemoteFactory.required_params(remote)
            missing_params = required_params - set(remote_config.keys())
            if missing_params:
                raise ValueError(
                    f"Missing required parameter(s) {missing_params} for remote {remote_name} in .trussrc"
                )

    @staticmethod
    def required_params(remote: Type[TrussRemote]) -> set:
        """
        Get the required parameters for a remote by inspecting its __init__ method
        """
        init_signature = inspect.signature(remote.__init__)
        params = init_signature.parameters
        required_params = {
            name
            for name, param in params.items()
            if param.default == inspect.Parameter.empty
            and name not in {"self", "args", "kwargs"}
        }
        return required_params

    @classmethod
    def create(cls, remote: str) -> TrussRemote:
        remote_config = cls.load_remote_config(remote)
        configs = remote_config.configs
        cls.validate_remote_config(configs, remote)

        remote_class = cls.REGISTRY[configs.pop("remote_provider")]
        remote_params = {
            param: configs.get(param) for param in cls.required_params(remote_class)
        }

        # Add any additional params provided by the user in their .trussrc
        additional_params = set(configs.keys()) - set(remote_params.keys())
        for param in additional_params:
            remote_params[param] = configs.get(param)

        return remote_class(**remote_params)  # type: ignore



import os
import signal
import socket
import subprocess
import threading
import time
from itertools import count

import briton_pb2
import briton_pb2_grpc
import grpc
from transformers import AutoTokenizer
from truss.config.trt_llm import TrussTRTLLMBuildConfiguration
from truss.constants import OPENAI_COMPATIBLE_TAG

BRITON_PORT = 50051

MODEL_INPUT_TO_BRITON_FIELD = {
    "max_tokens": "request_output_len",
    "beam_width": "beam_width",
    "repetition_penalty": "repetition_penalty",
    "presence_penalty": "presence_penalty",
    "temperature": "temperature",
    "length_penalty": "len_penalty",
    "end_id": "end_id",
    "pad_id": "pad_id",
    "runtime_top_k": "runtime_top_k",
    "runtime_top_p": "runtime_top_p",
}


def is_port_available(port, host="localhost"):
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            # Try to connect to the specified host and port
            s.bind((host, port))
            return True
    except OSError:
        # Port is not available
        return False


def briton_monitor(briton_process):
    while True:
        if briton_process.poll() is not None:
            print(
                f"Briton process has exited with code {briton_process.returncode}, exiting truss server"
            )
            pid = os.getpid()
            os.kill(pid, signal.SIGKILL)
        time.sleep(1)


class Model:
    def __init__(self, **kwargs):
        self._model = None
        self._config = kwargs["config"]
        self._data_dir = kwargs["data_dir"]
        self._stub = None
        self._secrets = kwargs["secrets"]
        self._request_id_counter = count(start=1)
        self._briton_process = None
        self._uses_openai_api = OPENAI_COMPATIBLE_TAG in self._config.get(
            "model_metadata", {}
        ).get("tags", [])

        if "trt_llm" not in self._config:
            raise ValueError("trt_llm config is required for this model")

        trtllm_config = self._config.get("trt_llm")
        truss_trtllm_build_config = TrussTRTLLMBuildConfiguration(
            **trtllm_config.get("build")
        )
        self._tp_count = truss_trtllm_build_config.tensor_parallel_count
        self._tokenizer_repository = (
            truss_trtllm_build_config.checkpoint_repository.repo
        )

        self._hf_token = None
        try:
            self._hf_token = self._secrets.get("hf_access_token", None)
        except:  # noqa
            pass

    def load(self):
        self._tokenizer = AutoTokenizer.from_pretrained(
            self._tokenizer_repository, token=self._hf_token
        )
        # Start engine
        config_str = f'engine_path: "{self._data_dir.resolve()}"\nhf_tokenizer: "{self._tokenizer_repository}"'
        config_pbtxt_path = (self._data_dir / "briton_config.pbtxt").resolve()
        config_pbtxt_path.write_text(config_str)
        briton_env = os.environ.copy()
        if self._hf_token is not None:
            briton_env["HF_ACCESS_TOKEN"] = self._hf_token
        if self._tp_count is None or self._tp_count == 1:
            self._briton_process = subprocess.Popen(
                ["Briton", "--config", str(config_pbtxt_path)], env=briton_env
            )
        else:
            self._briton_process = subprocess.Popen(
                [
                    "mpirun",
                    "--allow-run-as-root",
                    "-n",
                    f"{self._tp_count}",
                    "Briton",
                    "--config",
                    str(config_pbtxt_path),
                ],
                env=briton_env,
            )
        while is_port_available(BRITON_PORT):
            print("Waiting for Briton to start")
            time.sleep(1)

        briton_monitor_thread = threading.Thread(
            target=briton_monitor, args=(self._briton_process,)
        )
        briton_monitor_thread.start()

    async def predict(self, model_input):
        if self._stub is None:
            channel = grpc.aio.insecure_channel(f"localhost:{BRITON_PORT}")
            self._stub = briton_pb2_grpc.BritonStub(channel)

        prompt = model_input.get("prompt", None)
        if prompt is None and "messages" in model_input:
            messages = model_input.pop("messages")
            if self._uses_openai_api:
                prompt = self._tokenizer.apply_chat_template(messages, tokenize=False)

        if prompt is None or prompt == "":
            yield ""
            return
        request_id = int(str(os.getpid()) + str(next(self._request_id_counter)))
        request = briton_pb2.InferenceRequest(
            request_id=request_id,
            input_text=prompt,
        )
        # Set default end_id and pad_id
        if (
            hasattr(self._tokenizer, "eos_token_id")
            and self._tokenizer.eos_token_id is not None
        ):
            request.end_id = self._tokenizer.eos_token_id
        if (
            hasattr(self._tokenizer, "pad_token_id")
            and self._tokenizer.pad_token_id is not None
        ):
            request.pad_id = self._tokenizer.pad_token_id
        set_briton_request_fields_from_model_input(model_input, request)
        for words in ["bad_words", "stop_words"]:
            if words in model_input:
                for word in model_input[words].split(","):
                    getattr(request, words).append(word)

        try:
            async for response in self._stub.Infer(request):
                yield response.output_text
        except grpc.RpcError as ex:
            if ex.code() == grpc.StatusCode.INVALID_ARGUMENT:
                print(ex.details())
        except Exception as ex:
            print(f"An error has occurred: {ex}")
            raise ex


def set_briton_request_fields_from_model_input(model_input, briton_request):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            setattr(briton_request, briton_field, model_input_value)



"""
The `Model` class is an interface between the ML model that you're packaging and the model
server that you're running it on.

The main methods to implement here are:
* `load`: runs exactly once when the model server is spun up or patched and loads the
   model onto the model server. Include any logic for initializing your model, such
   as downloading model weights and loading the model into memory.
* `predict`: runs every time the model server is called. Include any logic for model
  inference and return the model output.

See https://truss.baseten.co/quickstart for more.
"""
import torch


class Model:
    def __init__(self, **kwargs):
        # Uncomment the following to get access
        # to various parts of the Truss config.

        # self._data_dir = kwargs["data_dir"]
        # self._config = kwargs["config"]
        # self._secrets = kwargs["secrets"]
        self._model = None

    def load(self):
        # Load model here and assign to self._model.
        pass

    def predict(self, model_input):
        # Run model inference here
        return torch.cuda.is_available()



import pathlib

import pydantic
from truss.templates.shared import secrets_resolver
from truss_chains import definitions

# Better: in >=3.10 use `TypeAlias`.
UserConfigT = pydantic.BaseModel


class TrussChainletModel:
    _context: definitions.DeploymentContext[UserConfigT]
    _chainlet: definitions.ABCChainlet

    def __init__(
        self, config: dict, data_dir: pathlib.Path, secrets: secrets_resolver.Secrets
    ) -> None:
        truss_metadata: definitions.TrussMetadata[
            UserConfigT
        ] = definitions.TrussMetadata[UserConfigT].model_validate(
            config["model_metadata"][definitions.TRUSS_CONFIG_CHAINS_KEY]
        )
        self._context = definitions.DeploymentContext[UserConfigT](
            user_config=truss_metadata.user_config,
            chainlet_to_service=truss_metadata.chainlet_to_service,
            secrets=secrets,
            data_dir=data_dir,
        )

    # Below illustrated code will be added by code generation.

    # def load(self) -> None:
    #     logging.info(f"Loading Chainlet `TextToNum`.")
    #     self._chainlet = main.TextToNum(
    #       mistral=stub.factory(MistralLLM, self._context))
    #
    # def predict(self, inputs: TextToNumInput) -> TextToNumOutput:
    #     with utils.exception_to_http_error(
    #         include_stack=True, chainlet_name="TextToNum"
    #     ):
    #         result = self._chainlet.run_remote(data=inputs.data)
    #     return TextToNumOutput((result,))



import warnings

import pydantic

pydantic_major_version = int(pydantic.VERSION.split(".")[0])
if pydantic_major_version < 2:
    warnings.warn(
        f"Pydantic major version is less than 2 ({pydantic.VERSION}). You use Truss, "
        "but for using Truss-Chains, you must upgrade to pydantic-v2.",
        UserWarning,
    )

del pydantic
del warnings

# flake8: noqa F401
from truss_chains.definitions import (
    Assets,
    ChainsRuntimeError,
    Compute,
    DeploymentContext,
    DockerImage,
    RemoteConfig,
    RemoteErrorDetail,
    RPCOptions,
)
from truss_chains.public_api import (
    ChainletBase,
    depends,
    depends_context,
    deploy_remotely,
    mark_entrypoint,
    run_local,
)
from truss_chains.stub import StubBase
from truss_chains.utils import make_abs_path_here



import builtins
import contextlib
import enum
import inspect
import logging
import os
import random
import socket
import sys
import textwrap
import traceback
from typing import Any, Iterable, Iterator, NoReturn, Type, TypeVar, Union

import fastapi
import httpx
import pydantic
from truss_chains import definitions

T = TypeVar("T")


def make_abs_path_here(file_path: str) -> definitions.AbsPath:
    """Helper to specify file paths relative to the *immediately calling* module.

    E.g. in you have a project structure like this

    root/
        chain.py
        common_requirements.text
        sub_package/
            Chainlet.py
            chainlet_requirements.txt

    Not in `root/sub_package/Chainlet.py` you can point to the requirements
    file like this:

    ```
    shared = RelativePathToHere("../common_requirements.text")
    specific = RelativePathToHere("chainlet_requirements.text")
    ```

    Caveat: this helper uses the directory of the immediately calling module as an
    absolute reference point for resolving the file location.
    Therefore, you MUST NOT wrap the instantiation of `RelativePathToHere` into a
    function (e.g. applying decorators) or use dynamic code execution.

    Ok:
    ```
    def foo(path: AbsPath):
        abs_path = path.abs_path


    foo(make_abs_path_here("blabla"))
    ```

    Not Ok:
    ```
    def foo(path: str):
        badbadbad = make_abs_path_here(path).abs_path

    foo("blabla"))
    ```
    """
    # TODO: the absolute path resolution below uses the calling module as a
    #   reference point. This would not work if users wrap this call in a function
    #   - we hope the naming makes clear that this should not be done.
    caller_frame = inspect.stack()[1]
    module_path = caller_frame.filename
    if not os.path.isabs(file_path):
        module_dir = os.path.dirname(os.path.abspath(module_path))
        abs_file_path = os.path.normpath(os.path.join(module_dir, file_path))
    else:
        abs_file_path = file_path

    return definitions.AbsPath(abs_file_path, module_path, file_path)


def setup_dev_logging(level: Union[int, str] = logging.INFO) -> None:
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    log_format = "%(levelname)s %(asctime)s %(filename)s:%(lineno)d] %(message)s"
    date_format = "%m%d %H:%M:%S"
    formatter = logging.Formatter(fmt=log_format, datefmt=date_format)
    if root_logger.handlers:
        for handler in root_logger.handlers:
            handler.setFormatter(formatter)
    else:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)


@contextlib.contextmanager
def log_level(level: int) -> Iterator[None]:
    """Change loglevel for code in this context."""
    current_logging_level = logging.getLogger().getEffectiveLevel()
    logging.getLogger().setLevel(level)
    try:
        yield
    finally:
        logging.getLogger().setLevel(current_logging_level)


def expect_one(it: Iterable[T]) -> T:
    """Assert that an iterable has exactly on element and return it."""
    it = iter(it)
    try:
        element = next(it)
    except StopIteration:
        raise ValueError("Iterable is empty.")

    try:
        _ = next(it)
    except StopIteration:
        return element

    raise ValueError("Iterable has more than one element.")


def get_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))  # Bind to a free port provided by the host.
        s.listen(1)  # Not necessary but included for completeness.
        port = s.getsockname()[1]  # Retrieve the port number assigned.
        return port


# Error Propagation Utils. #############################################################


def _handle_exception(
    exception: Exception, include_stack: bool, chainlet_name: str
) -> NoReturn:
    """Raises `fastapi.HTTPException` with `RemoteErrorDetail` as detail."""
    if hasattr(exception, "__module__"):
        exception_module_name = exception.__module__
    else:
        exception_module_name = None

    if include_stack:
        error_stack = traceback.extract_tb(exception.__traceback__)
        # Exclude the error handling functions from the stack trace.
        exclude_frames = {exception_to_http_error.__name__, handle_response.__name__}
        final_tb = [frame for frame in error_stack if frame.name not in exclude_frames]
        stack = list(
            [definitions.StackFrame.from_frame_summary(frame) for frame in final_tb]
        )
    else:
        stack = []

    error = definitions.RemoteErrorDetail(
        remote_name=chainlet_name,
        exception_cls_name=exception.__class__.__name__,
        exception_module_name=exception_module_name,
        exception_message=str(exception),
        user_stack_trace=stack,
    )
    raise fastapi.HTTPException(
        status_code=500, detail=error.model_dump()
    ) from exception


@contextlib.contextmanager
def exception_to_http_error(include_stack: bool, chainlet_name: str) -> Iterator[None]:
    try:
        yield
    except Exception as e:
        _handle_exception(e, include_stack, chainlet_name)


def _resolve_exception_class(
    error: definitions.RemoteErrorDetail,
) -> Type[Exception]:
    """Tries to find the exception class in builtins or imported libs,
    falls back to `definitions.GenericRemoteError` if not found."""
    exception_cls = None
    if error.exception_module_name is None:
        exception_cls = getattr(builtins, error.exception_cls_name, None)
    else:
        if mod := sys.modules.get(error.exception_module_name):
            exception_cls = getattr(mod, error.exception_cls_name, None)

    if exception_cls is None:
        logging.warning(
            f"Could not resolve exception with name `{error.exception_cls_name}` "
            f"and module `{error.exception_module_name}` - fall back to "
            f"`{definitions.GenericRemoteException.__name__}`."
        )
        exception_cls = definitions.GenericRemoteException

    return exception_cls


def handle_response(response: httpx.Response, remote_name: str) -> Any:
    """For successful requests returns JSON, otherwise raises error.

    If the response error contains `RemoteErrorDetail`, it tries to re-raise
    the same exception that was raised remotely and falls back to
    `GenericRemoteException` if the exception class could not be resolved.

    Exception messages are chained to trace back to the root cause, i.e. the first
    Chainlet that raised an exception. E.g. the message might look like this:

    ```
    RemoteChainletError in "Chain"
    Traceback (most recent call last):
      File "/app/model/Chainlet.py", line 112, in predict
        result = await self._chainlet.run(
      File "/app/model/Chainlet.py", line 79, in run
        value += self._text_to_num.run(part)
      File "/packages/remote_stubs.py", line 21, in run
        json_result = self._remote.predict_sync(json_args)
      File "/packages/truss_chains/stub.py", line 37, in predict_sync
        return utils.handle_response(
    ValueError: (showing remote errors, root message at the bottom)
    --> Preceding Remote Cause:
        RemoteChainletError in "TextToNum"
        Traceback (most recent call last):
          File "/app/model/Chainlet.py", line 113, in predict
            result = self._chainlet.run(data=payload["data"])
          File "/app/model/Chainlet.py", line 54, in run
            generated_text = self._replicator.run(data)
          File "/packages/remote_stubs.py", line 7, in run
            json_result = self._remote.predict_sync(json_args)
          File "/packages/truss_chains/stub.py", line 37, in predict_sync
            return utils.handle_response(
        ValueError: (showing remote errors, root message at the bottom)
        --> Preceding Remote Cause:
            RemoteChainletError in "TextReplicator"
            Traceback (most recent call last):
              File "/app/model/Chainlet.py", line 112, in predict
                result = self._chainlet.run(data=payload["data"])
              File "/app/model/Chainlet.py", line 36, in run
                raise ValueError(f"This input is too long: {len(data)}.")
            ValueError: This input is too long: 100.

    ```
    """
    if response.is_error:
        try:
            response_json = response.json()
        except Exception as e:
            raise ValueError(
                "Could not get JSON from error response. Status: "
                f"`{response.status_code}`."
            ) from e

        try:
            error_json = response_json["error"]
        except KeyError as e:
            logging.error(f"response.json(): {response_json}")
            raise ValueError(
                "Could not get `error` field from JSON from error response"
            ) from e

        try:
            error = definitions.RemoteErrorDetail.model_validate(error_json)
        except pydantic.ValidationError as e:
            if isinstance(error_json, str):
                msg = f"Remote error occurred in `{remote_name}`: '{error_json}'"
                raise definitions.GenericRemoteException(msg) from None
            raise ValueError(
                "Could not parse error. Error details are expected to be either a "
                "plain string (old truss models) or a serialized "
                f"`definitions.RemoteErrorDetail.__name__`, got:\n{repr(error_json)}"
            ) from e

        exception_cls = _resolve_exception_class(error)
        msg = (
            f"(showing remote errors, root message at the bottom)\n"
            f"--> Preceding Remote Cause:\n"
            f"{textwrap.indent(error.format(), '    ')}"
        )
        raise exception_cls(msg)

    return response.json()


class InjectedError(Exception):
    """Test error for debugging/dev."""


def random_fail(probability: float, msg: str):
    """Probabilistically raises `InjectedError` for debugging/dev."""
    if random.random() < probability:
        print(f"Random failure: {msg}")
        raise InjectedError(msg)


class StrEnum(str, enum.Enum):
    """
    Adapted from MIT-licensed
    https://github.com/irgeek/StrEnum/blob/master/strenum/__init__.py

    This is useful for Pydantic-based (de-)serialisation, as Pydantic takes the value
    of an enum member as the value to be (de-)serialised, and not the name of the
    member. With this, we can have the member name and value be the same by using
    `enum.auto()`.

    StrEnum is a Python `enum.Enum` that inherits from `str`. The `auto()` behavior
    uses the member name and lowers it. This is useful for compatibility with pydantic.
    Example usage:
    ```
    class Example(StrEnum):
        SOME_VALUE = enum.auto()
        ANOTHER_VALUE = enum.auto()
        TEST = enum.auto()

    assert Example.SOME_VALUE == "SOME_VALUE"
    assert Example.ANOTHER_VALUE.value == "ANOTHER_VALUE"
    assert Example.TEST.value == Example.TEST
    assert Example.TEST == Example("TEST")
    ```
    """

    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, str):
            raise TypeError(f"Values of StrEnums must be strings: Got `{repr(value)}`.")
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self) -> str:
        return str(self.value)

    def _generate_next_value_(name, *_) -> str:  # type: ignore[override]
        if name.upper() != name:
            raise ValueError(f"Python enum members should be upper case. Got `{name}`.")
        return name


def issubclass_safe(x: Any, cls: type) -> bool:
    """Like built-in `issubclass`, but works on non-type objects."""
    return isinstance(x, type) and issubclass(x, cls)



from pathlib import Path

# If changing model repo path, please updated inside tensorrt_llm config.pbtxt as well
TENSORRT_LLM_MODEL_REPOSITORY_PATH = Path("/packages/tensorrt_llm_model_repository/")
GRPC_SERVICE_PORT = 8001
HTTP_SERVICE_PORT = 8003
HF_AUTH_KEY_CONSTANT = "HUGGING_FACE_HUB_TOKEN"
TOKENIZER_KEY_CONSTANT = "TRITON_TOKENIZER_REPOSITORY"
ENTRYPOINT_MODEL_NAME = "ensemble"



import truss_chains as chains


class DummyGenerateData(chains.ChainletBase):
    def run_remote(self) -> str:
        return "abc"


# Nesting the classes is a hack to make it *appear* like SplitText is from a different
# module.
class shared_chainlet:
    class DummySplitText(chains.ChainletBase):
        def run_remote(self, data: str) -> list[str]:
            return [data[:2], data[2:]]


class DummyExample(chains.ChainletBase):
    def __init__(
        self,
        data_generator: DummyGenerateData = chains.depends(DummyGenerateData),
        splitter: shared_chainlet.DummySplitText = chains.depends(
            shared_chainlet.DummySplitText
        ),
        context: chains.DeploymentContext = chains.depends_context(),
    ) -> None:
        self._data_generator = data_generator
        self._data_splitter = splitter
        self._context = context

    def run_remote(self) -> list[str]:
        return self._data_splitter.run_remote(self._data_generator.run_remote())



import abc
import functools
import logging
from typing import Optional, Type, TypeVar, final

import httpx
import tenacity
from truss_chains import definitions, utils


class BasetenSession:
    """Helper to invoke predict method on baseten deployments."""

    def __init__(
        self,
        service_descriptor: definitions.ServiceDescriptor,
        api_key: str,
    ) -> None:
        logging.info(
            f"Creating BasetenSession (HTTP) for `{service_descriptor.name}` "
            f"({service_descriptor.options.retries} retries) with predict URL:\n"
            f"    `{service_descriptor.predict_url}`"
        )
        self._auth_header = {"Authorization": f"Api-Key {api_key}"}
        self._service_descriptor = service_descriptor

    @property
    def name(self) -> str:
        return self._service_descriptor.name

    @functools.cached_property
    def _client_sync(self) -> httpx.Client:
        return httpx.Client(
            headers=self._auth_header,
            timeout=self._service_descriptor.options.timeout_sec,
        )

    @functools.cached_property
    def _client_async(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(
            headers=self._auth_header,
            timeout=self._service_descriptor.options.timeout_sec,
        )

    def predict_sync(self, json_payload):
        retrying = tenacity.Retrying(
            stop=tenacity.stop_after_attempt(self._service_descriptor.options.retries),
            retry=tenacity.retry_if_exception_type(Exception),
            reraise=True,
        )
        for attempt in retrying:
            with attempt:
                if (num := attempt.retry_state.attempt_number) > 1:
                    exc = attempt.retry_state.outcome.exception()
                    logging.warning(
                        f"Retrying `{self.name}`: attempt {num} due to {exc}."
                    )
                return utils.handle_response(
                    self._client_sync.post(
                        self._service_descriptor.predict_url, json=json_payload
                    ),
                    self.name,
                )

    async def predict_async(self, json_payload):
        retrying = tenacity.AsyncRetrying(
            stop=tenacity.stop_after_attempt(self._service_descriptor.options.retries),
            retry=tenacity.retry_if_exception_type(Exception),
            reraise=True,
        )
        async for attempt in retrying:
            with attempt:
                if (num := attempt.retry_state.attempt_number) > 1:
                    exc = attempt.retry_state.outcome.exception()
                    logging.warning(
                        f"Retrying `{self.name}`: attempt {num} due to {exc}."
                    )
                return utils.handle_response(
                    await self._client_async.post(
                        self._service_descriptor.predict_url, json=json_payload
                    ),
                    self.name,
                )


class StubBase(abc.ABC):
    _remote: BasetenSession

    @final
    def __init__(
        self, service_descriptor: definitions.ServiceDescriptor, api_key: str
    ) -> None:
        self._remote = BasetenSession(service_descriptor, api_key)

    @classmethod
    def from_url(
        cls,
        predict_url: str,
        context: definitions.DeploymentContext,
        options: Optional[definitions.RPCOptions] = None,
        name: Optional[str] = None,
    ):
        name = name or cls.__name__
        options = options or definitions.RPCOptions()
        return cls(
            definitions.ServiceDescriptor(
                name=name, predict_url=predict_url, options=options
            ),
            api_key=context.get_baseten_api_key(),
        )


StubT = TypeVar("StubT", bound=StubBase)


def factory(stub_cls: Type[StubT], context: definitions.DeploymentContext) -> StubT:
    # Assumes the stub_cls-name and the name of the service in `context` match.
    return stub_cls(
        service_descriptor=context.get_service_descriptor(stub_cls.__name__),
        api_key=context.get_baseten_api_key(),
    )



# flake8: noqa
# type: ignore

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: briton.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0c\x62riton.proto\x12\x06\x62riton"\xb0\x04\n\x10InferenceRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x12\n\ninput_text\x18\x02 \x01(\t\x12\x11\n\tinput_ids\x18\x03 \x03(\x05\x12\x1f\n\x12request_output_len\x18\x05 \x01(\rH\x00\x88\x01\x01\x12\x13\n\x06\x65nd_id\x18\x06 \x01(\rH\x01\x88\x01\x01\x12\x13\n\x06pad_id\x18\x07 \x01(\rH\x02\x88\x01\x01\x12\x17\n\nbeam_width\x18\n \x01(\rH\x03\x88\x01\x01\x12\x18\n\x0btemperature\x18\x0b \x01(\x02H\x04\x88\x01\x01\x12\x1a\n\rruntime_top_k\x18\x0c \x01(\rH\x05\x88\x01\x01\x12\x1a\n\rruntime_top_p\x18\r \x01(\x02H\x06\x88\x01\x01\x12\x18\n\x0blen_penalty\x18\x0e \x01(\x02H\x07\x88\x01\x01\x12\x1f\n\x12repetition_penalty\x18\x0f \x01(\x02H\x08\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x10 \x01(\x02H\t\x88\x01\x01\x12\x11\n\tbad_words\x18\x11 \x03(\t\x12\x12\n\nstop_words\x18\x12 \x03(\tB\x15\n\x13_request_output_lenB\t\n\x07_end_idB\t\n\x07_pad_idB\r\n\x0b_beam_widthB\x0e\n\x0c_temperatureB\x10\n\x0e_runtime_top_kB\x10\n\x0e_runtime_top_pB\x0e\n\x0c_len_penaltyB\x15\n\x13_repetition_penaltyB\x13\n\x11_presence_penalty"R\n\x13InferenceAnswerPart\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x13\n\x0boutput_text\x18\x02 \x01(\t\x12\x12\n\noutput_ids\x18\x03 \x03(\x05"\xfa\x03\n\x0c\x42ritonConfig\x12\x13\n\x0b\x65ngine_path\x18\x01 \x01(\t\x12\x14\n\x0chf_tokenizer\x18\x02 \x01(\t\x12N\n\x16\x62\x61tch_scheduler_policy\x18\x05 \x01(\x0e\x32).briton.BritonConfig.BatchSchedulerPolicyH\x00\x88\x01\x01\x12\x1f\n\x12\x65nable_trt_overlap\x18\x06 \x01(\x08H\x01\x88\x01\x01\x12)\n\x1cmax_tokens_in_paged_kv_cache\x18\n \x01(\x04H\x02\x88\x01\x01\x12+\n\x1ekv_cache_free_gpu_mem_fraction\x18\x0b \x01(\x02H\x03\x88\x01\x01\x12!\n\x14medusa_decoding_mode\x18\x0c \x01(\x08H\x04\x88\x01\x01"D\n\x14\x42\x61tchSchedulerPolicy\x12\x13\n\x0fMAX_UTILIZATION\x10\x00\x12\x17\n\x13GUARANTEED_NO_EVICT\x10\x01\x42\x19\n\x17_batch_scheduler_policyB\x15\n\x13_enable_trt_overlapB\x1f\n\x1d_max_tokens_in_paged_kv_cacheB!\n\x1f_kv_cache_free_gpu_mem_fractionB\x17\n\x15_medusa_decoding_mode2L\n\x06\x42riton\x12\x42\n\x05Infer\x12\x18.briton.InferenceRequest\x1a\x1b.briton.InferenceAnswerPart"\x00\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "briton_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_INFERENCEREQUEST"]._serialized_start = 25
    _globals["_INFERENCEREQUEST"]._serialized_end = 585
    _globals["_INFERENCEANSWERPART"]._serialized_start = 587
    _globals["_INFERENCEANSWERPART"]._serialized_end = 669
    _globals["_BRITONCONFIG"]._serialized_start = 672
    _globals["_BRITONCONFIG"]._serialized_end = 1178
    _globals["_BRITONCONFIG_BATCHSCHEDULERPOLICY"]._serialized_start = 967
    _globals["_BRITONCONFIG_BATCHSCHEDULERPOLICY"]._serialized_end = 1035
    _globals["_BRITON"]._serialized_start = 1180
    _globals["_BRITON"]._serialized_end = 1256
# @@protoc_insertion_point(module_scope)



from dataclasses import replace
from pathlib import Path

import pytest
from truss.constants import SUPPORTED_PYTHON_VERSIONS
from truss.local.local_config_handler import LocalConfigHandler
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.tests.test_truss_handle import (
    verify_python_requirement_installed_on_container,
    verify_python_requirement_not_installed_on_container,
    verify_system_package_installed_on_container,
    verify_system_requirement_not_installed_on_container,
)
from truss.truss_config import ExternalDataItem
from truss.truss_gatherer import calc_shadow_truss_dirname
from truss.truss_handle import TrussHandle


def current_num_docker_images(th: TrussHandle) -> int:
    return len(th.get_all_docker_images())


@pytest.fixture
def control_model_handle_tag_tuple(
    custom_model_control,
) -> tuple[Path, TrussHandle, str]:
    th = TrussHandle(custom_model_control)
    tag = "test-docker-custom-model-control-tag:0.0.1"
    return (custom_model_control, th, tag)


def update_model_code(path, new_model_code):
    model_code_file_path = path / "model" / "model.py"
    with model_code_file_path.open("w") as model_code_file:
        model_code_file.write(new_model_code)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_model_code_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    new_model_code = """
class Model:
    def predict(self, model_input):
        return [2 for i in model_input]
"""

    with ensure_kill_all():
        result = th.docker_predict([1], tag=tag, binary=binary)
        assert result[0] == 1
        orig_num_truss_images = len(th.get_all_docker_images())
        update_model_code(custom_model_control, new_model_code)
        result = th.docker_predict([1], tag=tag, binary=binary)
        assert result[0] == 2
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_empty_dir_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_added_empty_directory():
        # Adding empty directory should work
        (custom_model_control / "model" / "dir").mkdir()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_added_empty_directory()
        assert orig_num_truss_images == current_num_docker_images(th)


# todo(justin): remove once this patch is supported
@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_unpatchable(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_unpatchable_change():
        # Changes that are not expressible with patch should also work
        # Changes to data dir are not currently patch expressible
        (custom_model_control / "data" / "dummy").touch()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_unpatchable_change()
        assert current_num_docker_images(th) == orig_num_truss_images + 1


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_python_sys_req_patch(
    binary, python_version, control_model_handle_tag_tuple
):
    _, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_python_requirement_added(req: str):
        th.add_python_requirement(req)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_python_requirement_removed(req):
        th.remove_python_requirement(req)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_system_requirement_added(pkg):
        th.add_system_package(pkg)
        return th.docker_predict([1], tag=tag, binary=binary)

    def predict_with_system_requirement_removed(pkg):
        th.remove_system_package(pkg)
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = len(th.get_all_docker_images())

        container = th.get_running_serving_container_ignore_hash()

        python_req = "pydot"
        predict_with_python_requirement_added(python_req)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_python_requirement_installed_on_container(container, python_req)

        predict_with_python_requirement_removed(python_req)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_python_requirement_not_installed_on_container(container, python_req)

        system_pkg = "jq"
        predict_with_system_requirement_added(system_pkg)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_system_package_installed_on_container(container, system_pkg)

        predict_with_system_requirement_removed(system_pkg)
        assert current_num_docker_images(th) == orig_num_truss_images
        verify_system_requirement_not_installed_on_container(container, system_pkg)


@pytest.mark.integration
@pytest.mark.parametrize(
    "binary, python_version",
    [
        (binary, python_version)
        for binary in [True, False]
        for python_version in SUPPORTED_PYTHON_VERSIONS
    ],
)
def test_control_truss_patch_ignored_changes(
    binary, python_version, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple
    th.update_python_version(python_version)

    def predict_with_ignored_changes():
        top_pycache_path = custom_model_control / "__pycache__"
        top_pycache_path.mkdir()
        (top_pycache_path / "bla.pyc").touch()
        model_pycache_path = custom_model_control / "model" / "__pycache__"
        model_pycache_path.mkdir()
        (model_pycache_path / "foo.pyc").touch()
        return th.docker_predict([1], tag=tag, binary=binary)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag, binary=binary)
        orig_num_truss_images = current_num_docker_images(th)

        predict_with_ignored_changes()
        assert current_num_docker_images(th) == orig_num_truss_images


@pytest.mark.integration
def test_patch_added_model_dir(
    custom_model_control, tmp_path, control_model_handle_tag_tuple
):
    custom_model_control, th, tag = control_model_handle_tag_tuple

    def predict_with_added_model_dir_file():
        code_file_dir = custom_model_control / "model" / "dir"
        code_file_dir.mkdir(parents=True)
        with (code_file_dir / "foo.bar").open("w") as model_code_file:
            model_code_file.write("foobar")
        return th.docker_predict([1], build_dir=tmp_path, tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_added_model_dir_file()
        assert orig_num_truss_images == current_num_docker_images(th)
        assert (custom_model_control / "model" / "dir" / "foo.bar").exists()


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_data_dir(control_model_handle_tag_tuple):
    custom_model_control, th, tag = control_model_handle_tag_tuple

    def predict_with_data_dir_change():
        path = custom_model_control / "data" / "dummy"
        path.touch()
        th.docker_predict([1], tag=tag)
        with path.open("w") as file:
            file.write("foobar")
        th.docker_predict([1], tag=tag)
        path.unlink()
        return th.docker_predict([1], tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_data_dir_change()
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
def test_patch_env_var(control_model_handle_tag_tuple):
    truss_dir, th, tag = control_model_handle_tag_tuple

    def predict_with_environment_variables_change():
        th.add_environment_variable("foo", "bar")
        result = th.docker_predict([1], tag=tag)
        assert result == "bar"
        th.add_environment_variable("foo", "bar2")
        result = th.docker_predict([1], tag=tag)
        assert result == "bar2"
        config_path = truss_dir / "config.yaml"
        with config_path.open("w") as file:
            file.write("{}")

        assert th._serving_hash() != th.truss_hash_on_serving_container()
        result = th.docker_predict([1], tag=tag)
        assert result == "No foo :("

    with ensure_kill_all():
        new_model_code = """
import os
class Model:
    def predict(self, model_input):
        print(os.environ.copy())
        return os.environ.get("foo", "No foo :(")
"""
        update_model_code(truss_dir, new_model_code)
        result = th.docker_predict([1], tag=tag)
        assert result == "No foo :("
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_environment_variables_change()
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
def test_patch_external_package_dirs(custom_model_with_external_package):
    th = TrussHandle(custom_model_with_external_package)
    tag = "test-docker-custom-model-control-external-package-tag:0.0.1"
    th.live_reload()
    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())
        th.clear_external_packages()
        th.add_external_package("../ext_pkg_patched")
        th.add_external_package("../ext_pkg2_patched")
        ext_pkg_path = custom_model_with_external_package / ".." / "ext_pkg_patched"
        ext_pkg_path.mkdir()
        (ext_pkg_path / "subdir_patched").mkdir()
        (ext_pkg_path / "subdir_patched" / "sub_module.py").touch()
        (ext_pkg_path / "top_module.py").touch()
        ext_pkg_path2 = custom_model_with_external_package / ".." / "ext_pkg2_patched"
        ext_pkg_path2.mkdir()
        (ext_pkg_path2 / "top_module_patched.py").touch()
        new_model_code = """
import top_module
import subdir_patched.sub_module
import top_module_patched
class Model:
    def predict(self, model_input):
        return [1 for i in model_input]
"""
        update_model_code(custom_model_with_external_package, new_model_code)
        th.docker_predict([1], tag=tag)
        assert orig_num_truss_images == current_num_docker_images(th)
        shadow_truss_path = (
            LocalConfigHandler.shadow_trusses_dir_path()
            / calc_shadow_truss_dirname(custom_model_with_external_package)
        )
        packages_dir_path_in_shadow = (
            shadow_truss_path / th.spec.config.bundled_packages_dir
        )
        assert (packages_dir_path_in_shadow / "subdir_patched").exists() and (
            packages_dir_path_in_shadow / "top_module_patched.py"
        ).exists()


# todo(abu/justin) remove once ignored
@pytest.mark.skip(reason="Unsupported patch")
@pytest.mark.integration
def test_patch_secrets(control_model_handle_tag_tuple):
    _, th, tag = control_model_handle_tag_tuple

    def predict_with_secrets():
        th.add_secret("foo", "bar")
        return th.docker_predict([1], tag=tag)

    with ensure_kill_all():
        th.docker_predict([1], tag=tag)
        orig_num_truss_images = len(th.get_all_docker_images())

        predict_with_secrets()
        assert orig_num_truss_images == current_num_docker_images(th)


@pytest.mark.integration
def test_predict_with_external_data_change(
    custom_model_external_data_access_tuple_fixture, tmp_path
):
    truss_dir, _ = custom_model_external_data_access_tuple_fixture
    th = TrussHandle(truss_dir)
    th.live_reload()
    tag = "test-external-data-access-tag:0.0.1"
    with ensure_kill_all():
        th.docker_predict([], tag=tag, network="host")
        orig_num_truss_images = len(th.get_all_docker_images())
        th.remove_all_external_data()
        assert th._serving_hash() != th.truss_hash_on_serving_container()
        new_model_code = """
class Model:
    def __init__(self, data_dir):
        self._data_dir = data_dir
        pass

    def predict(self, model_input):
        return None
"""
        update_model_code(truss_dir, new_model_code)
        th.docker_predict([], tag=tag, network="host")
        content = "foobar"
        filename = "foobar.txt"
        (tmp_path / filename).write_text(content)
        new_model_code = """
class Model:
    def __init__(self, data_dir):
        self._data_dir = data_dir
        pass

    def predict(self, model_input):
        return (self._data_dir / 'foobar.txt').read_text()
"""
        update_model_code(truss_dir, new_model_code)
        url = f"http://host.docker.internal:9089/{filename}"
        th.add_external_data_item(url, filename)
        result = th.docker_predict([], tag=tag, network="host")
        assert result == content

        content = "patched content"
        new_filename = "foobar-patched.txt"
        (tmp_path / new_filename).write_text(content)
        current_external_data = th._spec.config.external_data
        new_external_data = replace(
            current_external_data,
            items=[
                ExternalDataItem(
                    url=f"http://host.docker.internal:9089/{new_filename}",
                    local_data_path=filename,
                )
            ],
        )
        th._update_config(
            lambda conf: replace(
                conf,
                external_data=new_external_data,
            )
        )
        result = th.docker_predict([], tag=tag, network="host")
        assert result == content and orig_num_truss_images == current_num_docker_images(
            th
        )



# Copyright 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
from collections import OrderedDict

import numpy as np
import triton_python_backend_utils as pb_utils
from transformers import AutoTokenizer

# https://github.com/huggingface/tokenizers/blob/main/tokenizers/src/decoders/strip.rs#L8
INVALID_UNICODE_CHAR = "�"


class TritonPythonModel:
    """Your Python model must use the same class name. Every Python model
    that is created must have "TritonPythonModel" as the class name.
    """

    def initialize(self, args):
        """`initialize` is called only once when the model is being loaded.
        Implementing `initialize` function is optional. This function allows
        the model to initialize any state associated with this model.
        Parameters
        ----------
        args : dict
          Both keys and values are strings. The dictionary keys and values are:
          * model_config: A JSON string containing the model configuration
          * model_instance_kind: A string containing model instance kind
          * model_instance_device_id: A string containing model instance device ID
          * model_repository: Model repository path
          * model_version: Model version
          * model_name: Model name
        """
        # Parse model configs
        model_config = json.loads(args["model_config"])
        # NOTE: Keep this in sync with the truss model.py variable
        tokenizer_dir = os.environ["TRITON_TOKENIZER_REPOSITORY"]
        hf_auth_token = os.environ.get("HUGGING_FACE_HUB_TOKEN", None)

        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_dir,
            legacy=False,
            padding_side="left",
            trust_remote_code=True,
            token=hf_auth_token,
        )
        self.tokenizer.pad_token = self.tokenizer.eos_token

        # Parse model output configs
        output_config = pb_utils.get_output_config_by_name(model_config, "OUTPUT")
        # Convert Triton types to numpy types
        self.output_dtype = pb_utils.triton_string_to_numpy(output_config["data_type"])

        self.state_dict = OrderedDict()
        # TODO(pankaj) This should come from the batch size
        self.cache_size = 2048

    def execute(self, requests):
        """`execute` must be implemented in every Python model. `execute`
        function receives a list of pb_utils.InferenceRequest as the only
        argument. This function is called when an inference is requested
        for this model. Depending on the batching configuration (e.g. Dynamic
        Batching) used, `requests` may contain multiple requests. Every
        Python model, must create one pb_utils.InferenceResponse for every
        pb_utils.InferenceRequest in `requests`. If there is an error, you can
        set the error argument when creating a pb_utils.InferenceResponse.
        Parameters
        ----------
        requests : list
          A list of pb_utils.InferenceRequest
        Returns
        -------
        list
          A list of pb_utils.InferenceResponse. The length of this list must
          be the same as `requests`
        """

        responses = []

        # Every Python backend must iterate over everyone of the requests
        # and create a pb_utils.InferenceResponse for each of them.
        for idx, request in enumerate(requests):
            # Get request ID
            request_id = request.request_id()

            # Get input tensors
            tokens_batch = (
                pb_utils.get_input_tensor_by_name(request, "TOKENS_BATCH")
                .as_numpy()
                .flatten()
            )

            if len(tokens_batch) == 0:
                continue

            # Postprocess output data
            prev_token = self._get_var(request_id, "prev_token")
            token_buffer = self._get_var(request_id, "token_buffer")
            token_buffer = token_buffer if token_buffer is not None else []
            current_tokens = np.concatenate(
                (np.array(token_buffer, dtype=int), tokens_batch), dtype=int
            )
            current_tokens_decoded = self.tokenizer.decode(current_tokens)

            if len(current_tokens_decoded) == 0:
                responses.append(pb_utils.InferenceResponse())
                continue

            if current_tokens_decoded[-1] == INVALID_UNICODE_CHAR:
                # If the last token is invalid, we need to keep it in the buffer
                # for the next request to see if this is a multi-token unicode
                # character.
                self._store_var(request_id, "token_buffer", current_tokens)
                responses.append(pb_utils.InferenceResponse())
                continue

            if prev_token is None:
                delta = current_tokens_decoded
            else:
                # TODO(pankaj) Figure out how to make tokenizer.decode not
                # ignore initial whitespace so we can avoid this hack.
                # Get string with and without previous token and diff. This hack
                # is needed because tokenizer.decode strips initial whitespace.
                old_string = self.tokenizer.decode(prev_token)
                with_prev_token = np.concatenate((prev_token, current_tokens))
                new_string = self.tokenizer.decode(with_prev_token)
                delta = self._compute_delta(old_string, new_string)

            # The previous token is the last character of the decoded sequence
            # which includes the multi-token unicode character.
            self._store_var(request_id, "prev_token", current_tokens)
            self._store_var(request_id, "token_buffer", None)

            # Create output tensor
            output_tensor = pb_utils.Tensor(
                "OUTPUT", np.array([delta]).astype(self.output_dtype)
            )

            # Get cum log probs
            cum_log_probs = pb_utils.get_input_tensor_by_name(request, "CUM_LOG_PROBS")

            # Get sequence length
            output_log_probs = pb_utils.get_input_tensor_by_name(
                request, "OUTPUT_LOG_PROBS"
            )

            # Get context logits
            context_logits = pb_utils.get_input_tensor_by_name(
                request, "CONTEXT_LOGITS"
            )

            # Get generation logits
            generation_logits = pb_utils.get_input_tensor_by_name(
                request, "GENERATION_LOGITS"
            )

            outputs = []
            outputs.append(output_tensor)

            if cum_log_probs:
                out_cum_log_probs = pb_utils.Tensor(
                    "OUT_CUM_LOG_PROBS", cum_log_probs.as_numpy()
                )
                outputs.append(out_cum_log_probs)
            else:
                out_cum_log_probs = pb_utils.Tensor(
                    "OUT_CUM_LOG_PROBS", np.array([[0.0]], dtype=np.float32)
                )
                outputs.append(out_cum_log_probs)

            if output_log_probs:
                out_output_log_probs = pb_utils.Tensor(
                    "OUT_OUTPUT_LOG_PROBS", output_log_probs.as_numpy()
                )
                outputs.append(out_output_log_probs)
            else:
                out_output_log_probs = pb_utils.Tensor(
                    "OUT_OUTPUT_LOG_PROBS", np.array([[[0.0]]], dtype=np.float32)
                )
                outputs.append(out_output_log_probs)

            if context_logits:
                out_context_logits = pb_utils.Tensor(
                    "OUT_CONTEXT_LOGITS", context_logits.as_numpy()
                )
                outputs.append(out_context_logits)
            else:
                out_context_logits = pb_utils.Tensor(
                    "OUT_CONTEXT_LOGITS", np.array([[[0.0]]], dtype=np.float32)
                )
                outputs.append(out_context_logits)

            if generation_logits:
                out_generation_logits = pb_utils.Tensor(
                    "OUT_GENERATION_LOGITS", generation_logits.as_numpy()
                )
                outputs.append(out_generation_logits)
            else:
                out_generation_logits = pb_utils.Tensor(
                    "OUT_GENERATION_LOGITS", np.array([[[[0.0]]]], dtype=np.float32)
                )
                outputs.append(out_generation_logits)

            inference_response = pb_utils.InferenceResponse(output_tensors=outputs)
            responses.append(inference_response)

        return responses

    def finalize(self):
        print("Cleaning up...")

    def _store_var(self, request_id, var_name, var):
        if request_id in self.state_dict:
            self.state_dict[request_id][var_name] = var
            self.state_dict.move_to_end(request_id)
        else:
            if len(self.state_dict) > self.cache_size:
                self.state_dict.popitem(last=False)
            self.state_dict[request_id] = {"prev_token": None, "token_buffer": None}
            self.state_dict[request_id][var_name] = var

    def _get_var(self, request_id, var_name):
        if request_id in self.state_dict:
            return self.state_dict[request_id][var_name]
        return None

    def _compute_delta(self, prev_str, new_str):
        delta = "".join(
            [
                char
                for index, char in enumerate(new_str)
                if index >= len(prev_str) or char != prev_str[index]
            ]
        )
        return delta

    def _postprocessing(self, tokens):
        decoded_tokens = self.tokenizer.decode(tokens)
        return decoded_tokens



import functools
import pathlib
from typing import ContextManager, Mapping, Optional, Type, Union

from truss_chains import definitions, deploy, framework


def depends_context() -> definitions.DeploymentContext:
    """Sets a 'symbolic marker' for injecting a Context object at runtime.

    WARNING: Despite the type annotation, this does *not* immediately provide a
    context instance.
    Only when deploying remotely or using `run_local` a context instance is provided.
    """
    # The type error is silenced to because chains framework will at runtime inject
    # a corresponding instance. Nonetheless, we want to use a type annotation here,
    # to facilitate type inference, code-completion and type checking within the code
    # of chainlets that depend on the other chainlet.
    return framework.ContextDependencyMarker()  # type: ignore


def depends(
    chainlet_cls: Type[framework.ChainletT], retries: int = 1
) -> framework.ChainletT:
    """Sets a 'symbolic marker' for injecting a stub or local Chainlet at runtime.

    WARNING: Despite the type annotation, this does *not* immediately provide a
    chainlet instance.
    Only when deploying remotely or using `run_local` a chainlet instance is provided.
    """
    options = definitions.RPCOptions(retries=retries)
    # The type error is silenced to because chains framework will at runtime inject
    # a corresponding instance. Nonetheless, we want to use a type annotation here,
    # to facilitate type inference, code-completion and type checking within the code
    # of chainlets that depend on the other chainlet.
    return framework.ChainletDependencyMarker(chainlet_cls, options)  # type: ignore


class ChainletBase(definitions.ABCChainlet):
    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        framework.check_and_register_class(cls)
        # For default init (from `object`) we don't need to check anything.
        if cls.has_custom_init():
            original_init = cls.__init__

            @functools.wraps(original_init)
            def __init_with_arg_check__(self, *args, **kwargs):
                if args:
                    raise definitions.ChainsRuntimeError("Only kwargs are allowed.")
                framework.ensure_args_are_injected(cls, original_init, kwargs)
                original_init(self, *args, **kwargs)

            cls.__init__ = __init_with_arg_check__  # type: ignore[method-assign]


def mark_entrypoint(cls: Type[framework.ChainletT]) -> Type[framework.ChainletT]:
    return framework.entrypoint(cls)


def deploy_remotely(
    entrypoint: Type[definitions.ABCChainlet],
    chain_name: str,
    publish: bool = True,
    promote: bool = True,
    only_generate_trusses: bool = False,
) -> deploy.ChainService:
    options = definitions.DeploymentOptionsBaseten.create(
        chain_name=chain_name,
        publish=publish,
        promote=promote,
        only_generate_trusses=only_generate_trusses,
    )
    return deploy.deploy_remotely(entrypoint, options)


def run_local(
    secrets: Optional[Mapping[str, str]] = None,
    data_dir: Optional[Union[pathlib.Path, str]] = None,
    chainlet_to_service: Optional[Mapping[str, definitions.ServiceDescriptor]] = None,
) -> ContextManager[None]:
    """Context manager for using in-process instantiations of Chainlet dependencies."""
    data_dir = pathlib.Path(data_dir) if data_dir else None
    return framework.run_local(secrets, data_dir, chainlet_to_service)
